"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.plugin = exports.FunctionBuilder = void 0;
/**
 *
 * Copyright 2020 Tencent
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
const path_1 = __importDefault(require("path"));
const archiver_1 = __importDefault(require("archiver"));
const fs_1 = __importDefault(require("fs"));
const framework_core_1 = require("@cloudbase/framework-core");
const toolbox_1 = require("@cloudbase/toolbox");
class FunctionPlugin extends framework_core_1.Plugin {
    constructor(name, api, inputs) {
        super(name, api, inputs);
        this.name = name;
        this.api = api;
        this.inputs = inputs;
        const config = this.api.projectConfig;
        const DEFAULT_INPUTS = {
            functionRootPath: (config === null || config === void 0 ? void 0 : config.functionRoot) || 'cloudfunctions',
            functions: config === null || config === void 0 ? void 0 : config.functions,
            servicePaths: {},
        };
        this.resolvedInputs = resolveInputs(this.inputs, DEFAULT_INPUTS);
        this.resolvedInputs.functions = this.resolvedInputs.functions.map((func) => {
            return Object.assign({}, {
                runtime: 'Nodejs10.15',
                installDependency: true,
                handler: 'index.main',
            }, func);
        });
        this.functions = this.resolvedInputs.functions;
        this.functionRootPath = path_1.default.isAbsolute(this.resolvedInputs.functionRootPath)
            ? this.resolvedInputs.functionRootPath
            : path_1.default.join(this.api.projectPath, this.resolvedInputs.functionRootPath);
        this.builder = new FunctionBuilder({
            projectPath: this.api.projectPath,
        });
        this.outputs = {};
    }
    /**
     * 初始化
     */
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            this.api.logger.debug('FunctionPlugin: init', this.resolvedInputs);
        });
    }
    compile() {
        return __awaiter(this, void 0, void 0, function* () {
            this.api.logger.debug('FunctionPlugin: compile', this.resolvedInputs);
            const builderOptions = this.functions.map((func) => {
                var _a, _b;
                let fileName = func.name;
                let localFunctionPath;
                if ((_a = func.runtime) === null || _a === void 0 ? void 0 : _a.includes('Java')) {
                    fileName = func.name + '.jar';
                }
                if (func.functionDistPath) {
                    localFunctionPath = path_1.default.join(this.functionRootPath, func.functionDistPath);
                }
                else {
                    localFunctionPath = path_1.default.join(this.functionRootPath, fileName);
                }
                if (((_b = func.runtime) === null || _b === void 0 ? void 0 : _b.includes('Node')) && func.installDependency) {
                    const packageJSONExists = fs_1.default.existsSync(path_1.default.join(localFunctionPath, 'package.json'));
                    if (!packageJSONExists) {
                        this.api.logger.warn(`函数 ${func.name} 设置了云端安装依赖，但函数代码根目录下未提供 package.json`);
                        func.installDependency = false;
                    }
                }
                const zipName = `${func.name + Date.now()}.zip`;
                return {
                    name: func.name,
                    localPath: localFunctionPath,
                    zipFileName: zipName,
                    ignore: func.installDependency
                        ? ['node_modules/**/*', 'node_modules', ...(func.ignore || [])]
                        : [...(func.ignore || [])],
                };
            });
            const buildResult = yield this.builder.build(builderOptions);
            const codeUris = yield this.api.samManager.uploadFile(buildResult.functions.map((func) => {
                return {
                    fileType: 'FUNCTION',
                    fileName: `${func.name}.zip`,
                    filePath: func.source,
                };
            }));
            buildResult.functions.forEach((func, index) => {
                this.outputs[func.name] = codeUris[index];
            });
            return {
                EntryPoint: Object.values(this.resolvedInputs.servicePaths).map((servicePath) => {
                    return {
                        Label: '服务地址',
                        EntryType: 'HttpService',
                        HttpEntryPath: servicePath,
                    };
                }),
                Resources: this.functions.reduce((resources, func) => {
                    resources[this.toConstantCase(func.name)] = this.functionConfigToSAM(func);
                    return resources;
                }, {}),
            };
        });
    }
    /**
     * 执行本地命令
     */
    run() {
        return __awaiter(this, void 0, void 0, function* () { });
    }
    /**
     * 删除资源
     */
    remove() {
        return __awaiter(this, void 0, void 0, function* () { });
    }
    /**
     * 生成代码
     */
    genCode() {
        return __awaiter(this, void 0, void 0, function* () { });
    }
    /**
     * 构建
     */
    build() {
        return __awaiter(this, void 0, void 0, function* () {
            this.api.logger.debug('FunctionPlugin: build', this.resolvedInputs);
        });
    }
    /**
     * 部署
     */
    deploy() {
        return __awaiter(this, void 0, void 0, function* () {
            this.api.logger.debug('FunctionPlugin: deploy', this.resolvedInputs, this.buildOutput);
            // 批量部署云函数
            yield Promise.all(this.functions.map((func) => __awaiter(this, void 0, void 0, function* () {
                this.api.logger.info(`${this.api.emoji('🚀')} [${func.name}] 云函数部署成功`);
            })));
            this.api.logger.info(`${this.api.emoji('🚀')} 云函数部署成功`);
        });
    }
    functionConfigToSAM(functionConfig) {
        var _a;
        return Object.assign({
            Type: 'CloudBase::Function',
            Properties: Object.assign({
                Handler: functionConfig.handler || 'index.main',
                Description: 'CloudBase Framework 部署的云函数',
                Runtime: functionConfig.runtime,
                FunctionName: functionConfig.name,
                MemorySize: functionConfig.memory || 128,
                Timeout: functionConfig.timeout || 5,
                Environment: {
                    Variables: functionConfig.envVariables,
                },
                VpcConfig: functionConfig.vpc,
                HttpPath: this.resolvedInputs.servicePaths[functionConfig.name],
                InstallDependency: ((_a = functionConfig.runtime) === null || _a === void 0 ? void 0 : _a.includes('Node')) &&
                    'installDependency' in functionConfig
                    ? functionConfig.installDependency
                    : false,
                CodeUri: this.outputs[functionConfig.name] &&
                    this.outputs[functionConfig.name].codeUri,
                Role: 'TCB_QcsRole',
            }, this.api.bumpVersion && {
                NewVersion: true,
            }, this.api.versionRemark && {
                VersionRemark: this.api.versionRemark,
            }, functionConfig.aclRule && {
                AclTag: 'CUSTOM',
                AclRule: this.genAclRule(functionConfig),
            }),
        });
    }
    toConstantCase(name) {
        let result = '';
        let lastIsDivide = true;
        for (let i = 0; i < name.length; i++) {
            let letter = name[i];
            if (letter === '-' || letter === '_') {
                lastIsDivide = true;
            }
            else if (lastIsDivide) {
                result += letter.toUpperCase();
                lastIsDivide = false;
            }
            else {
                result += letter.toLowerCase();
                lastIsDivide = false;
            }
        }
        return result;
    }
    genAclRule(functionConfig) {
        const aclRule = {};
        aclRule[functionConfig.name] = functionConfig.aclRule;
        return JSON.stringify(aclRule);
    }
}
function resolveInputs(inputs, defaultInputs) {
    return Object.assign({}, defaultInputs, inputs);
}
class FunctionBuilder extends framework_core_1.Builder {
    constructor(options) {
        super(Object.assign({ type: 'function' }, options));
    }
    build(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const functions = yield Promise.all(options.map((option) => __awaiter(this, void 0, void 0, function* () {
                const localZipPath = path_1.default.join(this.distDir, option.zipFileName);
                if (!fs_1.default.existsSync(this.distDir)) {
                    toolbox_1.mkdirSync(this.distDir);
                }
                if (!fs_1.default.existsSync(option.localPath)) {
                    throw new Error(`函数目录或者文件 ${path_1.default.basename(option.localPath)} 不存在`);
                }
                const fileStats = fs_1.default.statSync(option.localPath);
                if (fileStats.isFile()) {
                    this.logger.debug('option.localPath', option.localPath, localZipPath);
                    yield this.zipFile(option.localPath, localZipPath);
                }
                else if (fileStats.isDirectory()) {
                    this.logger.debug('option.localPath', option.localPath, localZipPath);
                    yield this.zipDir(option.localPath, localZipPath, option.ignore);
                }
                return {
                    name: option.name,
                    options: {},
                    source: localZipPath,
                    entry: option.zipFileName,
                };
            })));
            return {
                functions,
            };
        });
    }
    zipFile(src, dest) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                // create a file to stream archive data to.
                var output = fs_1.default.createWriteStream(dest);
                var archive = archiver_1.default('zip', {
                    zlib: { level: 9 },
                });
                output.on('close', () => {
                    resolve();
                });
                archive.on('error', reject);
                archive.file(src, {
                    name: path_1.default.basename(src),
                });
                archive.pipe(output);
                archive.finalize();
            });
        });
    }
    zipDir(src, dest, ignore) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                // create a file to stream archive data to.
                var output = fs_1.default.createWriteStream(dest);
                var archive = archiver_1.default('zip', {
                    zlib: { level: 9 },
                });
                output.on('close', () => {
                    resolve();
                });
                archive.on('error', reject);
                archive.glob('**/*', {
                    cwd: src,
                    ignore: ignore || [],
                    dot: true,
                }, {});
                archive.pipe(output);
                archive.finalize();
            });
        });
    }
}
exports.FunctionBuilder = FunctionBuilder;
exports.plugin = FunctionPlugin;
